#!/usr/bin/env node

'use strict'

const { createServer } = require('http')
const { join } = require('path')
const { parse } = require('url')
const { watch } = require('fs')

const curContext = { version: '1.0', engine: 'lessev' }
const contentType = { 'Content-Type': 'application/json' }

let params = {
    port: 4000,
    entry: 'handler',
    method: 'POST',
    timeout: 3000,
    path: process.cwd(),
    mount: '/',
    limit: 2000
}

// parse options
process.argv.slice(2).forEach(str => {
    let [key, value] = str.split('=')
    params[key.replace(/^-+/, '')] = value
})

if (!['GET', 'POST'].includes(params.method)) {
    throw 'Method only support GET or POST'
}

// parse project main file
let metafilename = join(params.path, 'package.json')
let meta = require(metafilename)

// load project entry
let script = join(params.path, meta.main || 'index.js')
let scriptData = require(script)
let handler = scriptData[params.entry]

if (typeof handler !== 'function') throw 'Invalid entry'

// watch file change
let watcher = watch(params.path, { recursive: true }, () => {
    log('watch file change, reload module...')

    Object.keys(require.cache).forEach(i => {
        if (i !== __filename) delete require.cache[i]
    })

    scriptData = require(script)
    handler = scriptData[params.entry]
})

// start server
createServer(async (req, res) => {
    if (req.method !== params.method) {
        res.writeHead(404, contentType)
        return res.end(JSON.stringify({
            message: 'Not found'
        }))
    }

    let { query, pathname } = parse(req.url),
        called = false,
        now = Date.now()

    if (pathname !== params.mount) {
        res.writeHead(404, contentType)
        return res.end(JSON.stringify({
            message: 'Not found'
        }))
    }

    function callback(err, result) {
        if (called) return
        called = true
        log('Time used ' + (Date.now() - now))
        if (err) result = err instanceof Error ? err : new Error(err)
        if (result instanceof Error) {
            res.writeHead(500, contentType)
            res.end(JSON.stringify({ message: result.message }))
        } else {
            log('Response ' + JSON.stringify(result))
            res.writeHead(200, contentType)
            res.end(typeof result === 'object' ? JSON.stringify(result) : result)
        }
    }

    let event = await new Promise(resolve => {
        if (req.method === 'GET') {
            return resolve(query)
        } else {
            if (!(/json/.test(req.headers['content-type']))) {
                throw 'Invalid content-type'
            }
            req.setEncoding('utf8')
            let str = ''
            req.on('data', s => {
                str += s
                if (str.length > +params.limit) {
                    req.destroy()
                    return resolve(str)
                }
            })
            req.on('end', () => {
                try {
                    resolve(JSON.parse(str))
                } catch (err) {
                    log(err.message)
                    resolve(str)
                }
            })
        }
    })

    let context = Object.assign({
        date: new Date,
        timestamp: Date.now(),
        headers: req.headers,
        method: req.method
    }, curContext)

    log('Event ' + JSON.stringify(event))

    let result = handler(event, context, callback)

    setTimeout(() => !called && callback('Timeout'), +params.timeout)

    if (result instanceof Promise) {
        try {
            callback(null, await result)
        } catch (err) {
            !called && callback(err)
        }
    }
}).listen(+params.port, () => log('Port ' + params.port))

function log(msg) {
    console.log(
        '\x1b[38;5;34m' +
        new Date().toISOString() +
        '\x1b[0m : ' + msg
    )
}